//
//
// t44-morse.c
//
// 115200 baud FTDI character echo, with Morse code LED flashing
//
// set lfuse to 0x7E for 20 MHz xtal
//
// Based on code from:
// Neil Gershenfeld
// 12/8/10
//
// (c) Massachusetts Institute of Technology 2010
// Permission granted for experimental and personal use;
// license for commercial sale available from MIT.
//
// 
//

#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#define output(directions,pin) (directions |= pin) // set port direction for output
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 8.5 // bit delay for 115200 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define char_delay() _delay_ms(10) // char delay

#define serial_port PORTA
#define serial_direction DDRA
#define serial_pins PINA
#define serial_pin_in (1 << PA0)
#define serial_pin_out (1 << PA1)

#define led_delay() _delay_ms(200) // delay time between LED flashes

#define led_port PORTA
#define led_direction DDRA
#define led_pin (1 << PA3)

#define max_buffer 25

void get_char(volatile unsigned char *pins, unsigned char pin, char *rxbyte) {
   //
   // read character into rxbyte on pins pin
   //    assumes line driver (inverts bits)
   //
   *rxbyte = 0;
   while (pin_test(*pins,pin))
      //
      // wait for start bit
      //
      ;
   //
   // delay to middle of first data bit
   //
   half_bit_delay();
   bit_delay();
   //
   // unrolled loop to read data bits
   //
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 0);
   else
      *rxbyte |= (0 << 0);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 1);
   else
      *rxbyte |= (0 << 1);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 2);
   else
      *rxbyte |= (0 << 2);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 3);
   else
      *rxbyte |= (0 << 3);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 4);
   else
      *rxbyte |= (0 << 4);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 5);
   else
      *rxbyte |= (0 << 5);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 6);
   else
      *rxbyte |= (0 << 6);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 7);
   else
      *rxbyte |= (0 << 7);
   //
   // wait for stop bit
   //
   bit_delay();
   half_bit_delay();
   }

void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
   //
   // send character in txchar on port pin
   //    assumes line driver (inverts bits)
   //
   // start bit
   //
   clear(*port,pin);
   bit_delay();
   //
   // unrolled loop to write data bits
   //
   if bit_test(txchar,0)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,1)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,2)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,3)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,4)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,5)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,6)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,7)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   //
   // stop bit
   //
   set(*port,pin);
   bit_delay();
   //
   // char delay
   //
   bit_delay();
   }

void put_string(volatile unsigned char *port, unsigned char pin, char *str) {
   //
   // print a null-terminated string
   //
   static int index;
   index = 0;
   do {
      put_char(port, pin, str[index]);
      ++index;
      } while (str[index] != 0);
   }
   
void dot(){
	set(led_port, led_pin);
	led_delay(); // toggle LED on and off for a Morse code "dot"
	clear(led_port, led_pin);
	led_delay();
}

void dash(){
	set(led_port, led_pin);
	led_delay(); // toggle LED on for longer duration for Mose code "dash"
	led_delay();
	led_delay();
	clear(led_port, led_pin);
	led_delay();
}

void led_toggle(char *chr){
   switch(*chr){ // decide which character was typed and toggle LED accordingly
		case 'a':
			dot(); dash();
			break;
		case 'b':
			dash(); dot(); dot(); dot();
			break;
		case 'c':
			dash(); dot(); dash(); dot();
			break;
		case 'd':
			dash(); dot(); dot();
			break;
		case 'e':
			dot();
			break;
		case 'f':
			dot(); dot(); dash(); dot();
			break;
		case 'g':
			dash(); dash(); dot();
			break;
		case 'h':
			dot(); dot(); dot(); dot();
			break;
		case 'i':
			dot(); dot();
			break;
		case 'j':
			dot(); dash(); dash(); dash();
			break;
		case 'k':
			dash(); dot(); dash();
			break;
		case 'l':
			dot(); dash(); dot(); dot();
			break;
		case 'm':
			dash(); dash();
			break;
		case 'n':
			dash(); dot();
			break;
		case 'o':
			dash(); dash(); dash();
			break;
		case 'p':
			dot(); dash(); dash(); dot();
			break;
		case 'q':
			dash(); dash(); dot(); dash();
			break;
		case 'r':
			dot(); dash(); dot();
			break;
		case 's':
			dot(); dot(); dot();
			break;
		case 't':
			dash();
			break;
		case 'u':
			dot(); dot(); dash();
			break;
		case 'v':
			dot(); dot(); dot(); dash();
			break;
		case 'w':
			dot(); dash(); dash();
			break;
		case 'x':
			dash(); dot(); dot(); dash();
			break;
		case 'y':
			dash(); dot(); dash(); dash();
			break;
		case 'z':
			dash(); dash(); dot(); dot();
			break;
   }
}

int main(void) {
   //
   // main
   //
   static char chr;
   static char buffer[max_buffer] = {0};
   static int index;
   //
   // set clock divider to /1
   //
   CLKPR = (1 << CLKPCE);
   CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
   //
   // initialize output pins
   //
   set(serial_port, serial_pin_out);
   output(serial_direction, serial_pin_out);
   clear(led_port, led_pin); // initialize LED pins
   output(led_direction, led_pin);
   //
   // main loop
   //
   index = 0;
   while (1) {
      get_char(&serial_pins, serial_pin_in, &chr);
      put_string(&serial_port, serial_pin_out, "hello.ftdi.44.echo.c: you typed \"");
      buffer[index++] = chr;
	  if (index == (max_buffer-1))
         index = 0;
	  put_string(&serial_port, serial_pin_out, buffer);
      put_char(&serial_port, serial_pin_out, '\"');
      put_char(&serial_port, serial_pin_out, 10); // new line
	  led_toggle(&chr);
      }
   }
